from __future__ import annotations
from typing import Callable, Any

from paho.mqtt.client import Client
from simplemqtt import MQTTMessage

from simplemqtt.types import QualityOfService as QoS
from .mqtt_connection_base import MqttConnectionBase


class MQTTConnectionV3(MqttConnectionBase):
    def add_on_connect(self, on_connect: Callable[[MQTTConnectionV3, Client, Any, dict[str, int]], None]):
        """
        Register a user on_connect callback.
        Callback is invoked after a successful connect with (connection, client, userdata, flags).

        :param on_connect: Callable to execute after successful broker connection.
        :return: MqttBuilder
        """
        self._on_connect_callbacks.append(on_connect)

    def add_before_disconnect(self, before_disconnect: Callable[[MQTTConnectionV3], None]):
        """
        Register a user before_disconnect callback.
        Callback is invoked before connection is closed with (connection).

        :param before_disconnect: Callable to execute before disconnect.
        :return: MqttBuilder
        """
        self._before_disconnect_callbacks.append(before_disconnect)

    def add_on_disconnect(self, on_disconnect: Callable[[Client, Any, int], None]):
        """
        Register an on_disconnect callback.
        Callback is invoked before connection is closed with (connection).

        :param on_disconnect: Callable to execute after disconnect.
        """
        self._on_disconnect_callbacks.append(on_disconnect)

    def publish(self, topic: str, payload=None, qos: QoS = QoS.AtMostOnce,
                retain: bool = False, wait_for_publish: bool = False):
        """
        Publish a message.

        :param topic: Exact topic to publish to.
        :param payload: Bytes, str, or None. If None the broker receives a zero-length payload.
        :param qos: Delivery guarantee. 0=AtMostOnce, 1=AtLeastOnce, 2=ExactlyOnce.
        :param retain: If True, broker stores the last message for this topic and serves it to future subscribers.
        :param wait_for_publish: If True, block until the client reports the publishing is complete.

        :return:
        """
        return super()._publish(topic, payload, qos, retain, None, wait_for_publish)

    def subscribe(self, topic: str, on_message: Callable[[MQTTConnectionV3, Client, Any, MQTTMessage], None], qos: QoS = QoS.AtMostOnce):
        """
        Subscribe to a topic filter.

        :param topic: Topic filter to subscribe to. Wildcards '+' and '#' are allowed.
        :param on_message: Callback invoked for each matching message. Signature: (connection, client, userdata, msg) -> None.
        :param qos: Requested Quality of Service for this subscription. 0 = AtMostOnce, 1 = AtLeastOnce, 2 = ExactlyOnce.

        :return:
        """
        return super()._subscribe(topic, on_message, qos=qos)



from ..mqtt_connections import MQTTConnectionV3
from .mqtt_builder_base import MqttBuilder


class MQTTBuilderV3(MqttBuilder[MQTTConnectionV3]):
    def __init__(self, client_id: str, host: str):
        """
        Initialize the builder.

        :param client_id: Client identifier used by the MQTT client.
        :param host: Broker hostname or IPv4/IPv6 literal.
        """
        super().__init__(client_id, host, connector=MQTTConnectionV3)



# mqtt_message.py
from __future__ import annotations

import io
import json
import re
import paho.mqtt.client as mqtt
from functools import cached_property
from typing import Any, Optional, Union, Dict, Tuple
from simplemqtt import QualityOfService as QoS

try:
    from PIL import Image  # type: ignore
except Exception:
    Image = None  # type: ignore


StringOrBytes = Union[str, bytes, None]


def bytes_to_text(data: StringOrBytes, *, encoding: str) -> str:
    """Decode bytes to text with 'replace' and strip whitespace."""
    if data is None:
        return ""
    if isinstance(data, str):
        return data.strip()
    return data.decode(encoding, errors="replace").strip()


def parse_content_type(content_type_header: Optional[str]) -> Tuple[Optional[str], Dict[str, str]]:
    """
    Parse a HTTP-like Content-Type header into (media_type, parameters).
    Example: "application/json; charset=utf-8"
    """
    if not isinstance(content_type_header, str) or not content_type_header.strip():
        return None, {}
    segments = [segment.strip() for segment in content_type_header.split(";") if segment.strip()]
    media_type = segments[0].lower() if segments else None
    parameters: Dict[str, str] = {}
    for parameter in segments[1:]:
        if "=" in parameter:
            key, value = parameter.split("=", 1)
            parameters[key.strip().lower()] = value.strip().strip('"').strip("'")
    return media_type, parameters


def sniff_image_magic(payload_bytes: bytes) -> Optional[str]:
    """Tiny magic-number sniffing for common image formats."""
    if payload_bytes.startswith(b"\x89PNG\r\n\x1a\n"):
        return "image/png"
    if payload_bytes.startswith(b"\xff\xd8"):
        return "image/jpeg"
    if payload_bytes.startswith(b"GIF87a") or payload_bytes.startswith(b"GIF89a"):
        return "image/gif"
    if payload_bytes[:4] == b"RIFF" and payload_bytes[8:12] == b"WEBP":
        return "image/webp"
    if payload_bytes.startswith(b"BM"):
        return "image/bmp"
    head = payload_bytes[:256].lstrip()
    if head.startswith(b"<?xml") or head.startswith(b"<svg"):
        return "image/svg+xml"
    return None


class MQTTMessage:
    """Paho MQTTMessage wrapper with robust decoding, cached metadata, and type helpers."""

    # ---------- init ----------
    def __init__(self, message: mqtt.MQTTMessage):
        self._message = message
        self._default_text_encoding = "utf-8"

    # ---------- raw access (cheap; caching optional) ----------
    @property
    def raw(self) -> mqtt.MQTTMessage:
        return self._message

    @cached_property
    def topic(self) -> str:
        return getattr(self._message, "topic", "") or ""

    @cached_property
    def qos(self) -> Optional[QoS]:
        raw_value = getattr(self._message, "qos", None)
        if raw_value is None:
            return None
        try:
            return QoS(int(raw_value))
        except ValueError:
            return None

    @cached_property
    def retain(self) -> bool:
        return bool(getattr(self._message, "retain", False))

    @cached_property
    def _properties(self):
        return getattr(self._message, "properties", None)

    @cached_property
    def _payload_obj(self) -> StringOrBytes:
        return getattr(self._message, "payload", None)

    # ---------- v5 metadata ----------
    @cached_property
    def payload_format_indicator(self) -> Optional[int]:
        return getattr(self._properties, "PayloadFormatIndicator", None)

    @cached_property
    def content_type(self) -> Optional[str]:
        return getattr(self._properties, "ContentType", None)

    @cached_property
    def media_type_and_parameters(self) -> Tuple[Optional[str], Dict[str, str]]:
        return parse_content_type(self.content_type)

    @cached_property
    def media_type(self) -> Optional[str]:
        media_type, _ = self.media_type_and_parameters
        return media_type

    # ---------- encoding resolution ----------
    @cached_property
    def text_encoding(self) -> str:
        """
        Priority:
          1) MQTT v5 PayloadFormatIndicator == 1 => UTF-8
          2) Content-Type 'charset' parameter
          3) default_text_encoding
        """
        if self.payload_format_indicator == 1:
            return "utf-8"
        _, parameters = self.media_type_and_parameters
        charset = parameters.get("charset")
        if isinstance(charset, str) and charset:
            return charset
        return self._default_text_encoding

    # ---------- cached type flags ----------
    @cached_property
    def is_text(self) -> bool:
        if self.payload_format_indicator == 1:
            return True
        if not self.media_type:
            return False
        if self.media_type.startswith("text/"):
            return True
        if self.media_type in ("application/json", "application/xml"):
            return True
        if self.media_type.endswith("+json") or self.media_type.endswith("+xml"):
            return True
        if self.media_type == "image/svg+xml":
            return True
        return False

    @cached_property
    def is_json(self) -> bool:
        if self.media_type and (
            self.media_type == "application/json"
            or self.media_type.endswith("+json")
            or self.media_type == "text/json"
        ):
            return True
        try:
            _ = self.text  # will raise if undecodable
            json.loads(self.text)
            return True
        except Exception:
            return False

    @cached_property
    def is_image(self) -> bool:
        if self.media_type and (self.media_type.startswith("image/") or self.media_type == "image/svg+xml"):
            return True
        return sniff_image_magic(self.payload_bytes) is not None

    @cached_property
    def is_audio(self) -> bool:
        return bool(self.media_type and self.media_type.startswith("audio/"))

    @cached_property
    def is_binary(self) -> bool:
        return not self.is_text

    # ---------- cached conversions ----------
    @cached_property
    def payload_bytes(self) -> bytes:
        """Payload as bytes. Empty bytes if None."""
        payload = self._payload_obj
        if payload is None:
            return b""
        if isinstance(payload, (bytes, bytearray)):
            return bytes(payload)
        # Paho v3 may carry str; encode deterministically
        return bytes(str(payload), self.text_encoding)

    @cached_property
    def text(self) -> str:
        """Payload as text using resolved encoding. Always stripped."""
        return self.get_text(self.text_encoding)

    @cached_property
    def json_value(self) -> Any:
        """Parsed JSON value. Raises on invalid JSON."""
        return json.loads(self.text)

    @cached_property
    def image_bytes_and_media_type(self) -> Tuple[bytes, Optional[str]]:
        """
        Return (bytes, media_type) if the payload looks like an image.
        Raises TypeError if it does not.
        """
        media_type = self.media_type
        if not (media_type and media_type.startswith("image/")):
            sniffed = sniff_image_magic(self.payload_bytes)
            media_type = sniffed or media_type
        if not media_type or not (media_type.startswith("image/") or media_type == "image/svg+xml"):
            raise TypeError("Payload is not an image")
        return self.payload_bytes, media_type

    @cached_property
    def boolean_value(self) -> bool:
        """
        Convert payload to bool.
        Rules:
          - If JSON: return JSON boolean or numeric != 0 or string mapping.
          - Else text mapping (case-insensitive):
            true/false, on/off, yes/no, 1/0, open/closed, online/offline.
          - Fallback: non-empty string is True, empty is False.
        """
        if self.is_json:
            value = self.json_value
            if isinstance(value, bool):
                return value
            if isinstance(value, (int, float)):
                return value != 0
            text_value = value if isinstance(value, str) else str(value)
        else:
            text_value = self.text

        normalized = text_value.strip().lower()
        truthy = {"1", "true", "t", "yes", "y", "on", "open", "online"}
        falsy = {"0", "false", "f", "no", "n", "off", "closed", "offline"}
        if normalized in truthy:
            return True
        if normalized in falsy:
            return False
        return bool(normalized)

    @cached_property
    def integer_value(self) -> int:
        """
        Convert payload to int.
        - If JSON number -> int(value).
        - Else parse trimmed text strictly.
        """
        if self.is_json:
            value = self.json_value
            if isinstance(value, bool):
                return int(value)
            if isinstance(value, (int, float)):
                return int(value)
            text_value = str(value).strip()
        else:
            text_value = self.text

        candidate = text_value.strip()
        if not re.fullmatch(r"[+-]?\d+", candidate):
            raise ValueError(f"Payload is not an integer: {text_value!r}")
        return int(candidate)

    @cached_property
    def float_value(self) -> float:
        """
        Convert payload to float.
        - If JSON number -> float(value).
        - Else parse trimmed text strictly (supports scientific notation).
        """
        if self.is_json:
            value = self.json_value
            if isinstance(value, bool):
                return float(int(value))
            if isinstance(value, (int, float)):
                return float(value)
            text_value = str(value).strip()
        else:
            text_value = self.text

        candidate = text_value.strip()
        if not re.fullmatch(r"[+-]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?", candidate):
            raise ValueError(f"Payload is not a float: {text_value!r}")
        return float(candidate)

    # ---------- public getters (thin wrappers) ----------
    def get_image_pil(self):
        """Return a PIL.Image.Image. Requires Pillow to be installed."""
        if Image is None:
            raise ImportError("Pillow is not installed")
        image_bytes, _ = self.image_bytes_and_media_type
        return Image.open(io.BytesIO(image_bytes))

    def get_text(self, encoding) -> str:
        return bytes_to_text(self._payload_obj, encoding=encoding)

    # ---------- comparisons ----------
    def __str__(self) -> str:
        return self.text

    def __eq__(self, other: object) -> bool:
        """
        Supported comparisons:
          - str: decoded text equality
          - bytes/bytearray: raw bytes equality
          - dict/list: JSON deep equality
          - bool: boolean conversion equality
        """
        if isinstance(other, str):
            return self.text == other
        if isinstance(other, (bytes, bytearray)):
            other_bytes = other if isinstance(other, bytes) else bytes(other)
            return self.payload_bytes == other_bytes
        if isinstance(other, (dict, list)):
            try:
                return self.json_value == other
            except Exception:
                return False
        if isinstance(other, bool):
            try:
                return self.boolean_value is other
            except Exception:
                return False
        return NotImplemented

    def __repr__(self) -> str:
        preview = self.text if self.is_text else f"<{len(self.payload_bytes)} bytes>"
        return f"MQTTMessage(topic={self.topic!r}, media_type={self.media_type!r}, preview={preview!r})"



from __future__ import annotations

import paho.mqtt.client as mqtt
from paho.mqtt.properties import Properties
from typing import Callable, Any, Dict, Optional

from simplemqtt import MQTTMessage
from simplemqtt.setup_logging import get_logger
from simplemqtt.types import QualityOfService as QoS

logger = get_logger("MqttConnectionBase")


def get_rc(rc):
    rc = getattr(rc, "value", rc)
    success = rc == 0
    return success, rc


def invoke_callbacks(callbacks, callback_name, *args, **kwargs):
    for callback in callbacks:
        try:
            callback(*args, **kwargs)
        except Exception as e:
            logger.warning(f"{callback_name} handler error: {e}")


class MqttConnectionBase:
    def __init__(self):
        self._client = None
        self._connection_parameters = None
        self._availability_topic = None

        self._subscription_handlers: Dict[str, Callable[[Any, mqtt.Client, Any, MQTTMessage], None]] = {}
        self._on_connect_callbacks = []
        self._before_disconnect_callbacks = []
        self._on_disconnect_callbacks = []

    def inject_client(self, client: mqtt.Client, connection_parameters: dict, availability_topic: str | None) -> None:
        self._client: mqtt.Client = client
        self._client.on_connect = self._on_connect
        self._client.on_disconnect = self._on_disconnect
        self._client.on_message = self._on_message_handler
        self._connection_parameters = connection_parameters
        self._availability_topic = availability_topic

    @property
    def availability_topic(self) -> str | None:
        return self._availability_topic

    @property
    def is_connected(self):
        return self._client.is_connected

    def connect(self, blocking: bool = False, **connection_parameters):
        """
        Connect to Mqtt Broker.

        :param blocking: If True, run loop_forever(). If False, start a background network loop.
        :param connection_parameters: Kwargs to override connect parameters
        :return: MQTTConnectionV3 or MQTTConnectionV5 wrapper depending on protocol.
        """
        self._client.connect(**self._connection_parameters, **connection_parameters)

        if blocking:
            self._client.loop_forever()
        else:
            self._client.loop_start()

        return self

    def _version_filter(self, version3, version5):
        if self._client.protocol == mqtt.MQTTv5:
            return version5
        else:
            return version3

    def _on_connect_version_parameter_filter(self, client, userdata, flags, properties):
        return self._version_filter((self, client, userdata, flags), (self, client, userdata, flags, properties))

    def _on_disconnect_version_parameter_filter(self, client, userdata, rc, properties):
        return self._version_filter((client, userdata, rc), (client, userdata, rc, properties))

    def _on_connect(self, client, userdata, flags, rc, properties=None):
        success, rc = get_rc(rc)

        if success:
            logger.info("MQTT connected")
            invoke_callbacks(self._on_connect_callbacks, "On Connect",
                             *self._on_connect_version_parameter_filter(client, userdata, flags, properties))
        else:
            logger.error(f"MQTT connect failed rc={rc}")

    def _on_disconnect(self, client, userdata, rc, properties=None):
        success, rc = get_rc(rc)

        if success:
            logger.info("MQTT disconnected")
        else:
            rs = getattr(properties, "ReasonString", None)
            if rs:
                logger.warning(f"MQTT disconnected unexpectedly rc={rc} reason={rs}")
            else:
                logger.warning(f"MQTT disconnected unexpectedly rc={rc}")

        invoke_callbacks(self._on_disconnect_callbacks, "On Disconnect",
                         *self._on_disconnect_version_parameter_filter(client, userdata, rc, properties))

    def _on_message_handler(self, client, userdata, msg: mqtt.MQTTMessage):
        message = MQTTMessage(msg)
        on_message_callbacks = [handler for topic, handler in self._subscription_handlers.items()
                                if mqtt.topic_matches_sub(topic, message.topic)]
        invoke_callbacks(on_message_callbacks, f"On Message(Topic: {message.topic})", self, client, userdata, message)

    def _publish(self, topic: str, payload, qos: QoS = QoS.AtMostOnce, retain: bool = False,
                 properties: Optional[Properties] = None, wait_for_publish: bool = False) -> mqtt.MQTTMessageInfo:
        info = self._client.publish(topic, payload, qos, retain, properties)
        if wait_for_publish:
            info.wait_for_publish()
        return info

    def _subscribe(self, topic: str, on_message: Callable[[Any, mqtt.Client, Any, MQTTMessage], None], **kwargs) -> tuple[int, int]:
        self._subscription_handlers[topic] = on_message
        return self._client.subscribe(topic, **kwargs)

    def unsubscribe(self, *topics: str):
        """
        Remove local handlers and send UNSUBSCRIBE.

        :param topics: One or more topic filters to unsubscribe from.
        :return: None
        """
        topics = list(topics)
        for topic in topics:
            self._subscription_handlers.pop(topic, None)
        self._client.unsubscribe(topics)

    def close(self):
        """
        Stop network loop and disconnect cleanly.

        :return: None
        """
        invoke_callbacks(self._before_disconnect_callbacks, "Before Disconnect", self)
        self._client.loop_stop()
        self._client.disconnect()



import paho.mqtt.client as mqtt


class MQTTConfig:
    def __init__(self, client_id, host):
        self.client_id = client_id
        self.protocol = mqtt.MQTTv311
        self.clean_session = True

        self.host = host
        self.port = 1883
        self.keep_alive = 60

        self.username = None
        self.password = None

        self.auto_reconnect = None
        self.tls = None

        self.last_will = None
        self.on_connect_callbacks = []
        self.before_disconnect_callbacks = []
        self.on_disconnect_callbacks = []

    @property
    def require_login(self):
        return self.username is not None and self.password is not None

    @property
    def has_last_will(self):
        return self.last_will is not None

    @property
    def has_tls(self):
        return (self.tls is not None and
                all([key in self.tls for key in ("settings", "allow_insecure")]))

    @property
    def has_auto_reconnect(self):
        return (self.auto_reconnect is not None and
                all([key in self.auto_reconnect for key in ("min_delay", "max_delay")]))



from enum import IntEnum


class RetainHandling(IntEnum):
    # broker sends retained messages on every subscribe
    SendRetainedAlways = 0
    # broker sends retained messages only on a new subscription
    SendRetainedOnNewSubscription = 1
    # broker never sends retained messages for this subscription
    DoNotSendRetained = 2



from __future__ import annotations
from typing import Callable, Any, Optional

from paho.mqtt.client import Client
from paho.mqtt.properties import Properties
from paho.mqtt.subscribeoptions import SubscribeOptions

from simplemqtt import MQTTMessage
from simplemqtt.types import QualityOfService as QoS, RetainHandling
from .mqtt_connection_base import MqttConnectionBase


class MQTTConnectionV5(MqttConnectionBase):
    def add_on_connect(self, on_connect: Callable[[MQTTConnectionV5, Client, Any, dict[str, int], Optional[Properties]], None]):
        """
        Register an on_connect callback.
        Callback is invoked after a successful connect with (connection, client, userdata, flags, properties).

        :param on_connect: Callable to execute after successful broker connection.
        """
        self._on_connect_callbacks.append(on_connect)

    def add_before_disconnect(self, before_disconnect: Callable[[MQTTConnectionV5], None]):
        """
        Register a before_disconnect callback.
        Callback is invoked before connection is closed with (connection).

        :param before_disconnect: Callable to execute before disconnect.
        """
        self._before_disconnect_callbacks.append(before_disconnect)

    def add_on_disconnect(self, on_disconnect: Callable[[Client, Any, int, Optional[Properties]], None]):
        """
        Register an on_disconnect callback.
        Callback is invoked before connection is closed with (connection).

        :param on_disconnect: Callable to execute after disconnect.
        """
        self._on_disconnect_callbacks.append(on_disconnect)

    def publish(self, topic: str, payload, qos: QoS = QoS.AtMostOnce, retain: bool = False,
                properties: Optional[Properties] = None, wait_for_publish: bool = False):
        """
        Publish a message.

        :param topic: Exact topic to publish to.
        :param payload: Bytes, str, or None. If None the broker receives a zero-length payload.
        :param qos: Delivery guarantee. 0=AtMostOnce, 1=AtLeastOnce, 2=ExactlyOnce.
        :param retain: If True, broker stores the last message for this topic and serves it to future subscribers.
        :param properties: MQTT v5 properties for the PUBLISH packet.
        :param wait_for_publish: If True, block until the client reports the publishing is complete.

        :return: paho.mqtt.client.MQTTMessageInfo
        """
        return super()._publish(topic, payload, qos, retain, properties, wait_for_publish)

    def subscribe(self, topic: str, on_message: Callable[[MQTTConnectionV5, Client, Any, MQTTMessage], None], qos: QoS = QoS.AtMostOnce,
                  no_local: bool = False, retain_as_published: bool = False, retain_handling: RetainHandling = RetainHandling.SendRetainedAlways):
        """
        Subscribe to a topic filter with MQTT v5 subscribe options.

        :param topic: Topic filter to subscribe to. Wildcards '+' and '#' are allowed.
        :param on_message: Callback invoked for each matching message. Signature: (connection, client, userdata, msg) -> None.
        :param qos: Requested Quality of Service for this subscription. 0 = AtMostOnce, 1 = AtLeastOnce, 2 = ExactlyOnce.
        :param no_local: If True, the client will not receive its own published messages that match 'topic'.
            Useful for pub/sub loops. Comparison is done against the same client/session identity.
        :param retain_as_published: If True, the broker delivers the retained flag exactly as published.
            If False, the broker may clear or modify the retained flag on delivery.
            This allows consumers to distinguish retained messages via msg.retain.
        :param retain_handling:
            Controls when the broker sends retained messages upon subscribing.
            0 = Send retained always: Retained messages are delivered on every subscribe action.
            1 = Send retained on new subscription only: Retained messages are sent only when the subscription is newly created.
                Useful after reconnect with an existing session to avoid repeated retained floods.
            2 = Do not send retained: No retained messages are delivered at the time of subscribing.

        :return:
        """
        options = SubscribeOptions(qos, no_local, retain_as_published, retain_handling)
        return super()._subscribe(topic, on_message, options=options)




from .mqtt_connection_base import MqttConnectionBase
from .mqtt_connection_v3 import MQTTConnectionV3
from .mqtt_connection_v5 import MQTTConnectionV5

__all__ = ["MqttConnectionBase", "MQTTConnectionV3", "MQTTConnectionV5"]



"""Top-level package for simple-mqtt."""
from . import types
from .types import QualityOfService, RetainHandling
from .mqtt_message import MQTTMessage
from .mqtt_connections import MQTTConnectionV3, MQTTConnectionV5
from .mqtt_builder import MQTTBuilderV3, MQTTBuilderV5


__all__ = ["MQTTBuilderV3", "MQTTBuilderV5", "MQTTConnectionV3", "MQTTConnectionV5", "QualityOfService", "RetainHandling", "MQTTMessage"]
__version__ = "0.1.2"



from __future__ import annotations

from typing import TypeVar, Generic, Type

import paho.mqtt.client as mqtt
from paho.mqtt.packettypes import PacketTypes
from paho.mqtt.properties import Properties

from ..setup_logging import get_logger
from ..types import QualityOfService as QoS
from ..mqtt_config import MQTTConfig
from ..mqtt_connections import MQTTConnectionV3, MQTTConnectionV5


C = TypeVar("C", MQTTConnectionV3, MQTTConnectionV5)
logger = get_logger("MqttBuilder")


class MqttBuilder(Generic[C]):
    def __init__(self, client_id: str, host: str, connector: Type[C]):
        self._config = MQTTConfig(client_id, host)
        self._config.protocol = mqtt.MQTTv311 if connector is MQTTConnectionV3 else mqtt.MQTTv5

        self._connection: Type[C] = connector()

    def persistent_session(self, persistent_session: bool = True) -> MqttBuilder[C]:
        """
        Control session persistence.

        :param persistent_session: True for persistent session, False for clean start.
        :return: MqttBuilder
        """
        self._config.clean_session = not persistent_session
        return self

    def port(self, port: int) -> MqttBuilder[C]:
        """
        Set broker port. Default: 1883

        :param port: Usually 1883 (MQTT) or 8883 (MQTTS).
        :return: MqttBuilder
        """
        self._config.port = port
        return self

    def keep_alive(self, keep_alive: int) -> MqttBuilder[C]:
        """
        Set keepalive seconds.

        :param keep_alive: Interval in seconds for PINGREQ heartbeats.
        :return: MqttBuilder
        """
        self._config.keep_alive = keep_alive
        return self

    def login(self, username: str, password: str) -> MqttBuilder[C]:
        """
        Set username and password.

        :param username: Username string.
        :param password: Password string.
        :return: MqttBuilder
        """
        self._config.username = username
        self._config.password = password
        return self

    def availability(self, topic: str, payload_online: str = "online", payload_offline: str = "offline", qos: QoS = QoS.AtLeastOnce, retain: bool = True) -> MqttBuilder[C]:
        """
        Configure an availability topic.
        On successful connect, publish 'payload_online' to 'topic' with given qos/retain.
        Also sets the Last Will to 'payload_offline' for unclean disconnects.

        :param topic: Availability topic. Use a stable, retained topic so new subscribers see status immediately.
        :param payload_online: Payload published on connect.
        :param payload_offline: Payload set as Last Will and sent by broker on unclean disconnect.
        :param qos: Delivery level for availability messages. Usually 1.
        :param retain: Retain both online and will messages so late subscribers see the latest state.
        :return: MqttBuilder
        """
        args = topic, payload_online, qos, retain
        if isinstance(self._connection, MQTTConnectionV3):
            self._connection.add_on_connect(lambda connection, _1, _2, _3: connection.publish(*args))
        elif isinstance(self._connection, MQTTConnectionV5):
            self._connection.add_on_connect(lambda connection, _1, _2, _3, _4: connection.publish(*args))

        self._connection.add_before_disconnect(
            lambda connection: connection.publish(topic, payload_offline, qos, retain, wait_for_publish=True)
        )
        self.last_will(topic, payload_offline, qos, retain)
        return self

    def last_will(self, topic: str, payload: str = "offline",  qos: QoS = QoS.AtLeastOnce, retain: bool = True) -> MqttBuilder[C]:
        """
        Set MQTT Last Will and Testament.

        :param topic: Target topic for the will message.
        :param payload: Will payload sent by the broker on unclean disconnect.
        :param qos: Will QoS.
        :param retain: Whether the will is retained.
        :return: MqttBuilder
        """
        self._config.last_will = {
            "topic": topic,
            "payload": payload,
            "qos": qos,
            "retain": retain
        }
        return self

    # TLS Missing
    """    
    2. 
    client.tls_set(
        ca_certs="/path/ca.pem",
        certfile="/path/client.crt",
        keyfile="/path/client.key",
        # keyfile_password="***"  # optional if key encrypted
    )
    """

    def _tls(self, settings: dict, allow_insecure: bool = False) -> MqttBuilder[C]:
        self._config.tls = {
            "settings": settings,
            "allow_insecure": allow_insecure
        }
        return self

    def tls(self, allow_insecure: bool = False) -> MqttBuilder[C]:
        """
        Enable TLS with default settings.

        :param allow_insecure: If True, disable certificate hostname checks (insecure).
        :return: MqttBuilder
        """
        return self._tls({}, allow_insecure)

    def own_tls(self, ca_certs: str, allow_insecure: bool = False) -> MqttBuilder[C]:
        """
        Enable TLS with a custom CA bundle.

        :param ca_certs: Path to CA certificate bundle file.
        :param allow_insecure: If True, disable certificate hostname checks (insecure).
        :return: MqttBuilder
        """
        return self._tls({"ca_certs": ca_certs}, allow_insecure)

    def auto_reconnect(self, min_delay=1, max_delay=30) -> MqttBuilder[C]:
        """
        Enable exponential backoff reconnects.

        :param min_delay: Initial backoff in seconds before the first reconnect attempt.
        :param max_delay: Maximum backoff cap in seconds. The delay grows up to this value.
        :return: MqttBuilder
        """
        self._config.auto_reconnect = {
            "min_delay": min_delay,
            "max_delay": max_delay
        }
        return self

    def build(self, **additional_client_params) -> C:
        """
        Create the client and apply configuration.

        :param additional_client_params: Extra kwargs forwarded to paho.Client(...), e.g. transport="websockets".
        :return: MQTTConnectionV3 or MQTTConnectionV5 wrapper depending on protocol.
        """
        if self._config.protocol != mqtt.MQTTv5:
            additional_client_params["clean_session"] = self._config.clean_session

        client = mqtt.Client(client_id=self._config.client_id, protocol=self._config.protocol, **additional_client_params)

        if self._config.has_auto_reconnect:
            client.reconnect_delay_set(**self._config.auto_reconnect)

        if self._config.has_tls:
            client.tls_set(**self._config.tls["settings"])
            client.tls_insecure_set(self._config.tls["allow_insecure"])

        if self._config.require_login:
            client.username_pw_set(self._config.username, self._config.password)

        availability_topic = None
        if self._config.has_last_will:
            client.will_set(**self._config.last_will)
            availability_topic = self._config.last_will["topic"]

        connection_parameters = {
            "host": self._config.host,
            "port": self._config.port,
            "keepalive": self._config.keep_alive,
        }

        if self._config.protocol == mqtt.MQTTv5:
            # MQTT version 5 clean session. Is persistent for 3600 sec
            connection_parameters["clean_start"] = self._config.clean_session
            props = Properties(PacketTypes.CONNECT)
            props.SessionExpiryInterval = 3600 if not self._config.clean_session else 0  # seconds
            connection_parameters["properties"] = props

        self._connection.inject_client(client, connection_parameters, availability_topic)
        return self._connection

    def fast_build(self, **additional_client_params) -> C:
        """
        Create the client, apply configuration and connect.

        :param additional_client_params: Extra kwargs forwarded to paho.Client(...), e.g. transport="websockets".
        :return: MQTTConnectionV3 or MQTTConnectionV5 wrapper depending on protocol.
        """
        return self.build(**additional_client_params).connect()
